/* 
 * i-scream central monitoring system
 * http://www.i-scream.org.uk
 * Copyright (C) 2000-2002 i-scream
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

//---PACKAGE DECLARATION---
package uk.org.iscream.cms.util;

//---IMPORTS---
import java.net.ServerSocket;
import java.net.Socket;
import java.net.InetAddress;
import java.io.IOException;

/**
 * Access Control List ServerSocket wrapper. Used
 * to wrap a ServerSocket with an ACL, whilst extending
 * ServerSocket to make integration into code easier.
 * Once the ACL has been set, accept() can be used exactly
 * as it would be with a ServerSocket.
 *
 * @author  $Author: tdb $
 * @version $Id: ACLServerSocket.java,v 1.4 2003/02/05 14:27:58 tdb Exp $
 */
public class ACLServerSocket extends ServerSocket {

//---FINAL ATTRIBUTES---

    /**
     * The current CVS revision of this class
     */
    public static final String REVISION = "$Revision: 1.4 $";

//---STATIC METHODS---

//---CONSTRUCTORS---

    /**
     * See relevant ServerSocket constructor.
     */
    public ACLServerSocket(int port) throws IOException {
        super(port);
        setACL(new ACL());
    }
    
    /**
     * See relevant ServerSocket constructor.
     */
    public ACLServerSocket(int port, int backlog) throws IOException {
        super(port, backlog);
        setACL(new ACL());
    }
    
    /**
     * See relevant ServerSocket constructor.
     */
    public ACLServerSocket(int port, int backlog, InetAddress bindAddr) throws IOException {
        super(port, backlog, bindAddr);
        setACL(new ACL());
    }
    
    /**
     * See relevant ServerSocket constructor, and
     * sets the default acl.
     */
    public ACLServerSocket(ACL acl, int port) throws IOException {
        super(port);
        setACL(acl);
    }
    
    /**
     * See relevant ServerSocket constructor, and
     * sets the default acl.
     */
    public ACLServerSocket(ACL acl, int port, int backlog) throws IOException {
        super(port, backlog);
        setACL(acl);
    }
    
    /**
     * See relevant ServerSocket constructor, and
     * sets the default acl.
     */
    public ACLServerSocket(ACL acl, int port, int backlog, InetAddress bindAddr) throws IOException {
        super(port, backlog, bindAddr);
        setACL(acl);
    }

//---PUBLIC METHODS---

    /**
     * Essentially has the same behaviour as the
     * ServerSocket.accept() method, except won't
     * return a Socket unless it is permitted by
     * the ACL. Closes Socket's immediately that
     * aren't permitted by the ACL.
     */
    public Socket accept() throws IOException {
        while(true) {
            Socket s = super.accept();
            if(_acl.check(s.getInetAddress())) {
                return s;
            }
            else {
                s.close();
            }
        }
    }
    
    /**
     * Set the ACL at any point in the operation
     * of the ServerSocket.
     *
     * @param acl the new ACL to apply
     */
    public void setACL(ACL acl) {
        _acl = acl;
    }
    
    /**
     * Overrides the {@link java.lang.Object#toString() Object.toString()}
     * method to provide clean logging (every class should have this).
     *
     * This uses the uk.org.iscream.cms.server.util.FormatName class
     * to format the toString()
     *
     * @return the name of this class and its CVS revision
     */
    public String toString() {
        return FormatName.getName(
            _name,
            getClass().getName(),
            REVISION) + ":" + super.toString();
    }

//---PRIVATE METHODS---

//---ACCESSOR/MUTATOR METHODS---

//---ATTRIBUTES---

    /**
     * This is the friendly identifier of the
     * component this class is running in.
     * eg, a Filter may be called "filter1",
     * If this class does not have an owning
     * component,  a name from the configuration
     * can be placed here.  This name could also
     * be changed to null for utility classes.
     */
    private String _name = null;
    
    /**
     * The ACL used by this ACLServerSocket.
     */
    private ACL _acl;
    
//---STATIC ATTRIBUTES---

//---INNER CLASSES---

}
